{% extends "psynet_layout.html" %}

{% import "macros/prompt.html" as psynet_prompts %}
{% import "macros/control.html" as psynet_controls %}
{% import "macros/consent.html" as consent %}

{% macro next_button() %}
    <p class="vspace"></p>

    <button type="button" id="next-button" class="btn btn-primary btn-lg submit" onClick="psynet.nextPage();" style="margin-bottom: 30px">
        {{ gettext("Next") }}
    </button>

{% endmacro %}
